/*
 * unicode.c -- UTF-8 to UTF-32 conversions and various operations
 *
 * Copyright (c) 2013-2022 David Demelier <markand@malikania.fr>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <assert.h>
#include <errno.h>
#include <stdlib.h>

#include "unicode.h"

#define nelem(x) (sizeof (x) / sizeof *(x))

static int
cmp1(const void *v1, const void *v2)
{
	uint32_t r1 = *(uint32_t *)v1, r2 = *(uint32_t *)v2;

	return r1 - r2;
}

static int
cmp2(const void *v1, const void *v2)
{
	uint32_t r = *(uint32_t *)v1, *p = (uint32_t *)v2;

	if(r >= p[0] && r <= p[1])
		return 0;
	else
		return r - p[0];
}

static const uint32_t alpha3[][2] = {
	{ 0x00D6, 0x00D8 },
	{ 0x00F6, 0x00F8 },
	{ 0x02EC, 0x02EE },
	{ 0x0374, 0x0376 },
	{ 0x037D, 0x037F },
	{ 0x0386, 0x0388 },
	{ 0x038A, 0x038E },
	{ 0x03A1, 0x03A3 },
	{ 0x03F5, 0x03F7 },
	{ 0x052F, 0x0531 },
	{ 0x066F, 0x0671 },
	{ 0x06D3, 0x06D5 },
	{ 0x0710, 0x0712 },
	{ 0x0887, 0x0889 },
	{ 0x09A8, 0x09AA },
	{ 0x09B0, 0x09B2 },
	{ 0x09DD, 0x09DF },
	{ 0x0A28, 0x0A2A },
	{ 0x0A30, 0x0A32 },
	{ 0x0A33, 0x0A35 },
	{ 0x0A36, 0x0A38 },
	{ 0x0A5C, 0x0A5E },
	{ 0x0A8D, 0x0A8F },
	{ 0x0A91, 0x0A93 },
	{ 0x0AA8, 0x0AAA },
	{ 0x0AB0, 0x0AB2 },
	{ 0x0AB3, 0x0AB5 },
	{ 0x0B28, 0x0B2A },
	{ 0x0B30, 0x0B32 },
	{ 0x0B33, 0x0B35 },
	{ 0x0B5D, 0x0B5F },
	{ 0x0B83, 0x0B85 },
	{ 0x0B90, 0x0B92 },
	{ 0x0B9A, 0x0B9E },
	{ 0x0C0C, 0x0C0E },
	{ 0x0C10, 0x0C12 },
	{ 0x0C28, 0x0C2A },
	{ 0x0C8C, 0x0C8E },
	{ 0x0C90, 0x0C92 },
	{ 0x0CA8, 0x0CAA },
	{ 0x0CB3, 0x0CB5 },
	{ 0x0CDE, 0x0CE0 },
	{ 0x0D0C, 0x0D0E },
	{ 0x0D10, 0x0D12 },
	{ 0x0DB1, 0x0DB3 },
	{ 0x0DBB, 0x0DBD },
	{ 0x0E30, 0x0E32 },
	{ 0x0E82, 0x0E86 },
	{ 0x0E8A, 0x0E8C },
	{ 0x0EA3, 0x0EA7 },
	{ 0x0EB0, 0x0EB2 },
	{ 0x0EC4, 0x0EC6 },
	{ 0x0F47, 0x0F49 },
	{ 0x10C5, 0x10C7 },
	{ 0x10FA, 0x10FC },
	{ 0x1248, 0x124A },
	{ 0x1256, 0x125A },
	{ 0x1288, 0x128A },
	{ 0x12B0, 0x12B2 },
	{ 0x12BE, 0x12C2 },
	{ 0x12D6, 0x12D8 },
	{ 0x1310, 0x1312 },
	{ 0x167F, 0x1681 },
	{ 0x176C, 0x176E },
	{ 0x18A8, 0x18AA },
	{ 0x1CEC, 0x1CEE },
	{ 0x1CF3, 0x1CF5 },
	{ 0x1F57, 0x1F5F },
	{ 0x1FB4, 0x1FB6 },
	{ 0x1FBC, 0x1FBE },
	{ 0x1FC4, 0x1FC6 },
	{ 0x1FF4, 0x1FF6 },
	{ 0x2113, 0x2115 },
	{ 0x2124, 0x212A },
	{ 0x212D, 0x212F },
	{ 0x2D25, 0x2D27 },
	{ 0x2DA6, 0x2DA8 },
	{ 0x2DAE, 0x2DB0 },
	{ 0x2DB6, 0x2DB8 },
	{ 0x2DBE, 0x2DC0 },
	{ 0x2DC6, 0x2DC8 },
	{ 0x2DCE, 0x2DD0 },
	{ 0x2DD6, 0x2DD8 },
	{ 0x309F, 0x30A1 },
	{ 0x30FA, 0x30FC },
	{ 0x312F, 0x3131 },
	{ 0xA7D1, 0xA7D5 },
	{ 0xA801, 0xA803 },
	{ 0xA805, 0xA807 },
	{ 0xA80A, 0xA80C },
	{ 0xA8FB, 0xA8FD },
	{ 0xA9E4, 0xA9E6 },
	{ 0xA9FE, 0xAA00 },
	{ 0xAA42, 0xAA44 },
	{ 0xAAAF, 0xAAB1 },
	{ 0xAAC0, 0xAAC2 },
	{ 0xAB26, 0xAB28 },
	{ 0xAB2E, 0xAB30 },
	{ 0xAB5A, 0xAB5C },
	{ 0xFB1D, 0xFB1F },
	{ 0xFB28, 0xFB2A },
	{ 0xFB36, 0xFB38 },
	{ 0xFB3C, 0xFB40 },
	{ 0xFB41, 0xFB43 },
	{ 0xFB44, 0xFB46 },
	{ 0xFE74, 0xFE76 },
	{ 0x1000B, 0x1000D },
	{ 0x10026, 0x10028 },
	{ 0x1003A, 0x1003C },
	{ 0x1003D, 0x1003F },
	{ 0x10340, 0x10342 },
	{ 0x1057A, 0x1057C },
	{ 0x1058A, 0x1058C },
	{ 0x10592, 0x10594 },
	{ 0x10595, 0x10597 },
	{ 0x105A1, 0x105A3 },
	{ 0x105B1, 0x105B3 },
	{ 0x105B9, 0x105BB },
	{ 0x10785, 0x10787 },
	{ 0x107B0, 0x107B2 },
	{ 0x10808, 0x1080A },
	{ 0x10835, 0x10837 },
	{ 0x108F2, 0x108F4 },
	{ 0x10A13, 0x10A15 },
	{ 0x10A17, 0x10A19 },
	{ 0x10AC7, 0x10AC9 },
	{ 0x111DA, 0x111DC },
	{ 0x11211, 0x11213 },
	{ 0x11286, 0x1128A },
	{ 0x1128D, 0x1128F },
	{ 0x1129D, 0x1129F },
	{ 0x11328, 0x1132A },
	{ 0x11330, 0x11332 },
	{ 0x11333, 0x11335 },
	{ 0x114C5, 0x114C7 },
	{ 0x11913, 0x11915 },
	{ 0x11916, 0x11918 },
	{ 0x1193F, 0x11941 },
	{ 0x119E1, 0x119E3 },
	{ 0x11C08, 0x11C0A },
	{ 0x11D06, 0x11D08 },
	{ 0x11D09, 0x11D0B },
	{ 0x11D65, 0x11D67 },
	{ 0x11D68, 0x11D6A },
	{ 0x16FE1, 0x16FE3 },
	{ 0x1AFF3, 0x1AFF5 },
	{ 0x1AFFB, 0x1AFFD },
	{ 0x1AFFE, 0x1B000 },
	{ 0x1D454, 0x1D456 },
	{ 0x1D49C, 0x1D49E },
	{ 0x1D4AC, 0x1D4AE },
	{ 0x1D4B9, 0x1D4BD },
	{ 0x1D4C3, 0x1D4C5 },
	{ 0x1D505, 0x1D507 },
	{ 0x1D514, 0x1D516 },
	{ 0x1D51C, 0x1D51E },
	{ 0x1D539, 0x1D53B },
	{ 0x1D53E, 0x1D540 },
	{ 0x1D544, 0x1D546 },
	{ 0x1D550, 0x1D552 },
	{ 0x1D6C0, 0x1D6C2 },
	{ 0x1D6DA, 0x1D6DC },
	{ 0x1D6FA, 0x1D6FC },
	{ 0x1D714, 0x1D716 },
	{ 0x1D734, 0x1D736 },
	{ 0x1D74E, 0x1D750 },
	{ 0x1D76E, 0x1D770 },
	{ 0x1D788, 0x1D78A },
	{ 0x1D7A8, 0x1D7AA },
	{ 0x1D7C2, 0x1D7C4 },
	{ 0x1E7E6, 0x1E7E8 },
	{ 0x1E7EB, 0x1E7ED },
	{ 0x1E7EE, 0x1E7F0 },
	{ 0x1E7FE, 0x1E800 },
	{ 0x1EE03, 0x1EE05 },
	{ 0x1EE1F, 0x1EE21 },
	{ 0x1EE22, 0x1EE24 },
	{ 0x1EE27, 0x1EE29 },
	{ 0x1EE32, 0x1EE34 },
	{ 0x1EE37, 0x1EE3B },
	{ 0x1EE47, 0x1EE4D },
	{ 0x1EE4F, 0x1EE51 },
	{ 0x1EE52, 0x1EE54 },
	{ 0x1EE57, 0x1EE61 },
	{ 0x1EE62, 0x1EE64 },
	{ 0x1EE6A, 0x1EE6C },
	{ 0x1EE72, 0x1EE74 },
	{ 0x1EE77, 0x1EE79 },
	{ 0x1EE7C, 0x1EE80 },
	{ 0x1EE89, 0x1EE8B },
	{ 0x1EEA3, 0x1EEA5 },
	{ 0x1EEA9, 0x1EEAB },
};

static const uint32_t alpha2[][2] = {
	{ 0x0041, 0x005A },
	{ 0x0061, 0x007A },
	{ 0x00C0, 0x00D6 },
	{ 0x00D8, 0x00F6 },
	{ 0x00F8, 0x02C1 },
	{ 0x02C6, 0x02D1 },
	{ 0x02E0, 0x02E4 },
	{ 0x0370, 0x0374 },
	{ 0x0376, 0x0377 },
	{ 0x037A, 0x037D },
	{ 0x0388, 0x038A },
	{ 0x038E, 0x03A1 },
	{ 0x03A3, 0x03F5 },
	{ 0x03F7, 0x0481 },
	{ 0x048A, 0x052F },
	{ 0x0531, 0x0556 },
	{ 0x0560, 0x0588 },
	{ 0x05D0, 0x05EA },
	{ 0x05EF, 0x05F2 },
	{ 0x0620, 0x064A },
	{ 0x066E, 0x066F },
	{ 0x0671, 0x06D3 },
	{ 0x06E5, 0x06E6 },
	{ 0x06EE, 0x06EF },
	{ 0x06FA, 0x06FC },
	{ 0x0712, 0x072F },
	{ 0x074D, 0x07A5 },
	{ 0x07CA, 0x07EA },
	{ 0x07F4, 0x07F5 },
	{ 0x0800, 0x0815 },
	{ 0x0840, 0x0858 },
	{ 0x0860, 0x086A },
	{ 0x0870, 0x0887 },
	{ 0x0889, 0x088E },
	{ 0x08A0, 0x08C9 },
	{ 0x0904, 0x0939 },
	{ 0x0958, 0x0961 },
	{ 0x0971, 0x0980 },
	{ 0x0985, 0x098C },
	{ 0x098F, 0x0990 },
	{ 0x0993, 0x09A8 },
	{ 0x09AA, 0x09B0 },
	{ 0x09B6, 0x09B9 },
	{ 0x09DC, 0x09DD },
	{ 0x09DF, 0x09E1 },
	{ 0x09F0, 0x09F1 },
	{ 0x0A05, 0x0A0A },
	{ 0x0A0F, 0x0A10 },
	{ 0x0A13, 0x0A28 },
	{ 0x0A2A, 0x0A30 },
	{ 0x0A32, 0x0A33 },
	{ 0x0A35, 0x0A36 },
	{ 0x0A38, 0x0A39 },
	{ 0x0A59, 0x0A5C },
	{ 0x0A72, 0x0A74 },
	{ 0x0A85, 0x0A8D },
	{ 0x0A8F, 0x0A91 },
	{ 0x0A93, 0x0AA8 },
	{ 0x0AAA, 0x0AB0 },
	{ 0x0AB2, 0x0AB3 },
	{ 0x0AB5, 0x0AB9 },
	{ 0x0AE0, 0x0AE1 },
	{ 0x0B05, 0x0B0C },
	{ 0x0B0F, 0x0B10 },
	{ 0x0B13, 0x0B28 },
	{ 0x0B2A, 0x0B30 },
	{ 0x0B32, 0x0B33 },
	{ 0x0B35, 0x0B39 },
	{ 0x0B5C, 0x0B5D },
	{ 0x0B5F, 0x0B61 },
	{ 0x0B85, 0x0B8A },
	{ 0x0B8E, 0x0B90 },
	{ 0x0B92, 0x0B95 },
	{ 0x0B99, 0x0B9A },
	{ 0x0B9E, 0x0B9F },
	{ 0x0BA3, 0x0BA4 },
	{ 0x0BA8, 0x0BAA },
	{ 0x0BAE, 0x0BB9 },
	{ 0x0C05, 0x0C0C },
	{ 0x0C0E, 0x0C10 },
	{ 0x0C12, 0x0C28 },
	{ 0x0C2A, 0x0C39 },
	{ 0x0C58, 0x0C5A },
	{ 0x0C60, 0x0C61 },
	{ 0x0C85, 0x0C8C },
	{ 0x0C8E, 0x0C90 },
	{ 0x0C92, 0x0CA8 },
	{ 0x0CAA, 0x0CB3 },
	{ 0x0CB5, 0x0CB9 },
	{ 0x0CDD, 0x0CDE },
	{ 0x0CE0, 0x0CE1 },
	{ 0x0CF1, 0x0CF2 },
	{ 0x0D04, 0x0D0C },
	{ 0x0D0E, 0x0D10 },
	{ 0x0D12, 0x0D3A },
	{ 0x0D54, 0x0D56 },
	{ 0x0D5F, 0x0D61 },
	{ 0x0D7A, 0x0D7F },
	{ 0x0D85, 0x0D96 },
	{ 0x0D9A, 0x0DB1 },
	{ 0x0DB3, 0x0DBB },
	{ 0x0DC0, 0x0DC6 },
	{ 0x0E01, 0x0E30 },
	{ 0x0E32, 0x0E33 },
	{ 0x0E40, 0x0E46 },
	{ 0x0E81, 0x0E82 },
	{ 0x0E86, 0x0E8A },
	{ 0x0E8C, 0x0EA3 },
	{ 0x0EA7, 0x0EB0 },
	{ 0x0EB2, 0x0EB3 },
	{ 0x0EC0, 0x0EC4 },
	{ 0x0EDC, 0x0EDF },
	{ 0x0F40, 0x0F47 },
	{ 0x0F49, 0x0F6C },
	{ 0x0F88, 0x0F8C },
	{ 0x1000, 0x102A },
	{ 0x1050, 0x1055 },
	{ 0x105A, 0x105D },
	{ 0x1065, 0x1066 },
	{ 0x106E, 0x1070 },
	{ 0x1075, 0x1081 },
	{ 0x10A0, 0x10C5 },
	{ 0x10D0, 0x10FA },
	{ 0x10FC, 0x1248 },
	{ 0x124A, 0x124D },
	{ 0x1250, 0x1256 },
	{ 0x125A, 0x125D },
	{ 0x1260, 0x1288 },
	{ 0x128A, 0x128D },
	{ 0x1290, 0x12B0 },
	{ 0x12B2, 0x12B5 },
	{ 0x12B8, 0x12BE },
	{ 0x12C2, 0x12C5 },
	{ 0x12C8, 0x12D6 },
	{ 0x12D8, 0x1310 },
	{ 0x1312, 0x1315 },
	{ 0x1318, 0x135A },
	{ 0x1380, 0x138F },
	{ 0x13A0, 0x13F5 },
	{ 0x13F8, 0x13FD },
	{ 0x1401, 0x166C },
	{ 0x166F, 0x167F },
	{ 0x1681, 0x169A },
	{ 0x16A0, 0x16EA },
	{ 0x16F1, 0x16F8 },
	{ 0x1700, 0x1711 },
	{ 0x171F, 0x1731 },
	{ 0x1740, 0x1751 },
	{ 0x1760, 0x176C },
	{ 0x176E, 0x1770 },
	{ 0x1780, 0x17B3 },
	{ 0x1820, 0x1878 },
	{ 0x1880, 0x1884 },
	{ 0x1887, 0x18A8 },
	{ 0x18B0, 0x18F5 },
	{ 0x1900, 0x191E },
	{ 0x1950, 0x196D },
	{ 0x1970, 0x1974 },
	{ 0x1980, 0x19AB },
	{ 0x19B0, 0x19C9 },
	{ 0x1A00, 0x1A16 },
	{ 0x1A20, 0x1A54 },
	{ 0x1B05, 0x1B33 },
	{ 0x1B45, 0x1B4C },
	{ 0x1B83, 0x1BA0 },
	{ 0x1BAE, 0x1BAF },
	{ 0x1BBA, 0x1BE5 },
	{ 0x1C00, 0x1C23 },
	{ 0x1C4D, 0x1C4F },
	{ 0x1C5A, 0x1C7D },
	{ 0x1C80, 0x1C88 },
	{ 0x1C90, 0x1CBA },
	{ 0x1CBD, 0x1CBF },
	{ 0x1CE9, 0x1CEC },
	{ 0x1CEE, 0x1CF3 },
	{ 0x1CF5, 0x1CF6 },
	{ 0x1D00, 0x1DBF },
	{ 0x1E00, 0x1F15 },
	{ 0x1F18, 0x1F1D },
	{ 0x1F20, 0x1F45 },
	{ 0x1F48, 0x1F4D },
	{ 0x1F50, 0x1F57 },
	{ 0x1F5F, 0x1F7D },
	{ 0x1F80, 0x1FB4 },
	{ 0x1FB6, 0x1FBC },
	{ 0x1FC2, 0x1FC4 },
	{ 0x1FC6, 0x1FCC },
	{ 0x1FD0, 0x1FD3 },
	{ 0x1FD6, 0x1FDB },
	{ 0x1FE0, 0x1FEC },
	{ 0x1FF2, 0x1FF4 },
	{ 0x1FF6, 0x1FFC },
	{ 0x2090, 0x209C },
	{ 0x210A, 0x2113 },
	{ 0x2119, 0x211D },
	{ 0x212A, 0x212D },
	{ 0x212F, 0x2139 },
	{ 0x213C, 0x213F },
	{ 0x2145, 0x2149 },
	{ 0x2183, 0x2184 },
	{ 0x2C00, 0x2CE4 },
	{ 0x2CEB, 0x2CEE },
	{ 0x2CF2, 0x2CF3 },
	{ 0x2D00, 0x2D25 },
	{ 0x2D30, 0x2D67 },
	{ 0x2D80, 0x2D96 },
	{ 0x2DA0, 0x2DA6 },
	{ 0x2DA8, 0x2DAE },
	{ 0x2DB0, 0x2DB6 },
	{ 0x2DB8, 0x2DBE },
	{ 0x2DC0, 0x2DC6 },
	{ 0x2DC8, 0x2DCE },
	{ 0x2DD0, 0x2DD6 },
	{ 0x2DD8, 0x2DDE },
	{ 0x3005, 0x3006 },
	{ 0x3031, 0x3035 },
	{ 0x303B, 0x303C },
	{ 0x3041, 0x3096 },
	{ 0x309D, 0x309F },
	{ 0x30A1, 0x30FA },
	{ 0x30FC, 0x30FF },
	{ 0x3105, 0x312F },
	{ 0x3131, 0x318E },
	{ 0x31A0, 0x31BF },
	{ 0x31F0, 0x31FF },
	{ 0x9FFF, 0xA48C },
	{ 0xA4D0, 0xA4FD },
	{ 0xA500, 0xA60C },
	{ 0xA610, 0xA61F },
	{ 0xA62A, 0xA62B },
	{ 0xA640, 0xA66E },
	{ 0xA67F, 0xA69D },
	{ 0xA6A0, 0xA6E5 },
	{ 0xA717, 0xA71F },
	{ 0xA722, 0xA788 },
	{ 0xA78B, 0xA7CA },
	{ 0xA7D0, 0xA7D1 },
	{ 0xA7D5, 0xA7D9 },
	{ 0xA7F2, 0xA801 },
	{ 0xA803, 0xA805 },
	{ 0xA807, 0xA80A },
	{ 0xA80C, 0xA822 },
	{ 0xA840, 0xA873 },
	{ 0xA882, 0xA8B3 },
	{ 0xA8F2, 0xA8F7 },
	{ 0xA8FD, 0xA8FE },
	{ 0xA90A, 0xA925 },
	{ 0xA930, 0xA946 },
	{ 0xA960, 0xA97C },
	{ 0xA984, 0xA9B2 },
	{ 0xA9E0, 0xA9E4 },
	{ 0xA9E6, 0xA9EF },
	{ 0xA9FA, 0xA9FE },
	{ 0xAA00, 0xAA28 },
	{ 0xAA40, 0xAA42 },
	{ 0xAA44, 0xAA4B },
	{ 0xAA60, 0xAA76 },
	{ 0xAA7E, 0xAAAF },
	{ 0xAAB5, 0xAAB6 },
	{ 0xAAB9, 0xAABD },
	{ 0xAADB, 0xAADD },
	{ 0xAAE0, 0xAAEA },
	{ 0xAAF2, 0xAAF4 },
	{ 0xAB01, 0xAB06 },
	{ 0xAB09, 0xAB0E },
	{ 0xAB11, 0xAB16 },
	{ 0xAB20, 0xAB26 },
	{ 0xAB28, 0xAB2E },
	{ 0xAB30, 0xAB5A },
	{ 0xAB5C, 0xAB69 },
	{ 0xAB70, 0xABE2 },
	{ 0xD7B0, 0xD7C6 },
	{ 0xD7CB, 0xD7FB },
	{ 0xF900, 0xFA6D },
	{ 0xFA70, 0xFAD9 },
	{ 0xFB00, 0xFB06 },
	{ 0xFB13, 0xFB17 },
	{ 0xFB1F, 0xFB28 },
	{ 0xFB2A, 0xFB36 },
	{ 0xFB38, 0xFB3C },
	{ 0xFB40, 0xFB41 },
	{ 0xFB43, 0xFB44 },
	{ 0xFB46, 0xFBB1 },
	{ 0xFBD3, 0xFD3D },
	{ 0xFD50, 0xFD8F },
	{ 0xFD92, 0xFDC7 },
	{ 0xFDF0, 0xFDFB },
	{ 0xFE70, 0xFE74 },
	{ 0xFE76, 0xFEFC },
	{ 0xFF21, 0xFF3A },
	{ 0xFF41, 0xFF5A },
	{ 0xFF66, 0xFFBE },
	{ 0xFFC2, 0xFFC7 },
	{ 0xFFCA, 0xFFCF },
	{ 0xFFD2, 0xFFD7 },
	{ 0xFFDA, 0xFFDC },
	{ 0x10000, 0x1000B },
	{ 0x1000D, 0x10026 },
	{ 0x10028, 0x1003A },
	{ 0x1003C, 0x1003D },
	{ 0x1003F, 0x1004D },
	{ 0x10050, 0x1005D },
	{ 0x10080, 0x100FA },
	{ 0x10280, 0x1029C },
	{ 0x102A0, 0x102D0 },
	{ 0x10300, 0x1031F },
	{ 0x1032D, 0x10340 },
	{ 0x10342, 0x10349 },
	{ 0x10350, 0x10375 },
	{ 0x10380, 0x1039D },
	{ 0x103A0, 0x103C3 },
	{ 0x103C8, 0x103CF },
	{ 0x10400, 0x1049D },
	{ 0x104B0, 0x104D3 },
	{ 0x104D8, 0x104FB },
	{ 0x10500, 0x10527 },
	{ 0x10530, 0x10563 },
	{ 0x10570, 0x1057A },
	{ 0x1057C, 0x1058A },
	{ 0x1058C, 0x10592 },
	{ 0x10594, 0x10595 },
	{ 0x10597, 0x105A1 },
	{ 0x105A3, 0x105B1 },
	{ 0x105B3, 0x105B9 },
	{ 0x105BB, 0x105BC },
	{ 0x10600, 0x10736 },
	{ 0x10740, 0x10755 },
	{ 0x10760, 0x10767 },
	{ 0x10780, 0x10785 },
	{ 0x10787, 0x107B0 },
	{ 0x107B2, 0x107BA },
	{ 0x10800, 0x10805 },
	{ 0x1080A, 0x10835 },
	{ 0x10837, 0x10838 },
	{ 0x1083F, 0x10855 },
	{ 0x10860, 0x10876 },
	{ 0x10880, 0x1089E },
	{ 0x108E0, 0x108F2 },
	{ 0x108F4, 0x108F5 },
	{ 0x10900, 0x10915 },
	{ 0x10920, 0x10939 },
	{ 0x10980, 0x109B7 },
	{ 0x109BE, 0x109BF },
	{ 0x10A10, 0x10A13 },
	{ 0x10A15, 0x10A17 },
	{ 0x10A19, 0x10A35 },
	{ 0x10A60, 0x10A7C },
	{ 0x10A80, 0x10A9C },
	{ 0x10AC0, 0x10AC7 },
	{ 0x10AC9, 0x10AE4 },
	{ 0x10B00, 0x10B35 },
	{ 0x10B40, 0x10B55 },
	{ 0x10B60, 0x10B72 },
	{ 0x10B80, 0x10B91 },
	{ 0x10C00, 0x10C48 },
	{ 0x10C80, 0x10CB2 },
	{ 0x10CC0, 0x10CF2 },
	{ 0x10D00, 0x10D23 },
	{ 0x10E80, 0x10EA9 },
	{ 0x10EB0, 0x10EB1 },
	{ 0x10F00, 0x10F1C },
	{ 0x10F30, 0x10F45 },
	{ 0x10F70, 0x10F81 },
	{ 0x10FB0, 0x10FC4 },
	{ 0x10FE0, 0x10FF6 },
	{ 0x11003, 0x11037 },
	{ 0x11071, 0x11072 },
	{ 0x11083, 0x110AF },
	{ 0x110D0, 0x110E8 },
	{ 0x11103, 0x11126 },
	{ 0x11150, 0x11172 },
	{ 0x11183, 0x111B2 },
	{ 0x111C1, 0x111C4 },
	{ 0x11200, 0x11211 },
	{ 0x11213, 0x1122B },
	{ 0x11280, 0x11286 },
	{ 0x1128A, 0x1128D },
	{ 0x1128F, 0x1129D },
	{ 0x1129F, 0x112A8 },
	{ 0x112B0, 0x112DE },
	{ 0x11305, 0x1130C },
	{ 0x1130F, 0x11310 },
	{ 0x11313, 0x11328 },
	{ 0x1132A, 0x11330 },
	{ 0x11332, 0x11333 },
	{ 0x11335, 0x11339 },
	{ 0x1135D, 0x11361 },
	{ 0x11400, 0x11434 },
	{ 0x11447, 0x1144A },
	{ 0x1145F, 0x11461 },
	{ 0x11480, 0x114AF },
	{ 0x114C4, 0x114C5 },
	{ 0x11580, 0x115AE },
	{ 0x115D8, 0x115DB },
	{ 0x11600, 0x1162F },
	{ 0x11680, 0x116AA },
	{ 0x11700, 0x1171A },
	{ 0x11740, 0x11746 },
	{ 0x11800, 0x1182B },
	{ 0x118A0, 0x118DF },
	{ 0x118FF, 0x11906 },
	{ 0x1190C, 0x11913 },
	{ 0x11915, 0x11916 },
	{ 0x11918, 0x1192F },
	{ 0x119A0, 0x119A7 },
	{ 0x119AA, 0x119D0 },
	{ 0x11A0B, 0x11A32 },
	{ 0x11A5C, 0x11A89 },
	{ 0x11AB0, 0x11AF8 },
	{ 0x11C00, 0x11C08 },
	{ 0x11C0A, 0x11C2E },
	{ 0x11C72, 0x11C8F },
	{ 0x11D00, 0x11D06 },
	{ 0x11D08, 0x11D09 },
	{ 0x11D0B, 0x11D30 },
	{ 0x11D60, 0x11D65 },
	{ 0x11D67, 0x11D68 },
	{ 0x11D6A, 0x11D89 },
	{ 0x11EE0, 0x11EF2 },
	{ 0x12000, 0x12399 },
	{ 0x12480, 0x12543 },
	{ 0x12F90, 0x12FF0 },
	{ 0x13000, 0x1342E },
	{ 0x14400, 0x14646 },
	{ 0x16800, 0x16A38 },
	{ 0x16A40, 0x16A5E },
	{ 0x16A70, 0x16ABE },
	{ 0x16AD0, 0x16AED },
	{ 0x16B00, 0x16B2F },
	{ 0x16B40, 0x16B43 },
	{ 0x16B63, 0x16B77 },
	{ 0x16B7D, 0x16B8F },
	{ 0x16E40, 0x16E7F },
	{ 0x16F00, 0x16F4A },
	{ 0x16F93, 0x16F9F },
	{ 0x16FE0, 0x16FE1 },
	{ 0x18800, 0x18CD5 },
	{ 0x1AFF0, 0x1AFF3 },
	{ 0x1AFF5, 0x1AFFB },
	{ 0x1AFFD, 0x1AFFE },
	{ 0x1B000, 0x1B122 },
	{ 0x1B150, 0x1B152 },
	{ 0x1B164, 0x1B167 },
	{ 0x1B170, 0x1B2FB },
	{ 0x1BC00, 0x1BC6A },
	{ 0x1BC70, 0x1BC7C },
	{ 0x1BC80, 0x1BC88 },
	{ 0x1BC90, 0x1BC99 },
	{ 0x1D400, 0x1D454 },
	{ 0x1D456, 0x1D49C },
	{ 0x1D49E, 0x1D49F },
	{ 0x1D4A5, 0x1D4A6 },
	{ 0x1D4A9, 0x1D4AC },
	{ 0x1D4AE, 0x1D4B9 },
	{ 0x1D4BD, 0x1D4C3 },
	{ 0x1D4C5, 0x1D505 },
	{ 0x1D507, 0x1D50A },
	{ 0x1D50D, 0x1D514 },
	{ 0x1D516, 0x1D51C },
	{ 0x1D51E, 0x1D539 },
	{ 0x1D53B, 0x1D53E },
	{ 0x1D540, 0x1D544 },
	{ 0x1D54A, 0x1D550 },
	{ 0x1D552, 0x1D6A5 },
	{ 0x1D6A8, 0x1D6C0 },
	{ 0x1D6C2, 0x1D6DA },
	{ 0x1D6DC, 0x1D6FA },
	{ 0x1D6FC, 0x1D714 },
	{ 0x1D716, 0x1D734 },
	{ 0x1D736, 0x1D74E },
	{ 0x1D750, 0x1D76E },
	{ 0x1D770, 0x1D788 },
	{ 0x1D78A, 0x1D7A8 },
	{ 0x1D7AA, 0x1D7C2 },
	{ 0x1D7C4, 0x1D7CB },
	{ 0x1DF00, 0x1DF1E },
	{ 0x1E100, 0x1E12C },
	{ 0x1E137, 0x1E13D },
	{ 0x1E290, 0x1E2AD },
	{ 0x1E2C0, 0x1E2EB },
	{ 0x1E7E0, 0x1E7E6 },
	{ 0x1E7E8, 0x1E7EB },
	{ 0x1E7ED, 0x1E7EE },
	{ 0x1E7F0, 0x1E7FE },
	{ 0x1E800, 0x1E8C4 },
	{ 0x1E900, 0x1E943 },
	{ 0x1EE00, 0x1EE03 },
	{ 0x1EE05, 0x1EE1F },
	{ 0x1EE21, 0x1EE22 },
	{ 0x1EE29, 0x1EE32 },
	{ 0x1EE34, 0x1EE37 },
	{ 0x1EE4D, 0x1EE4F },
	{ 0x1EE51, 0x1EE52 },
	{ 0x1EE61, 0x1EE62 },
	{ 0x1EE67, 0x1EE6A },
	{ 0x1EE6C, 0x1EE72 },
	{ 0x1EE74, 0x1EE77 },
	{ 0x1EE79, 0x1EE7C },
	{ 0x1EE80, 0x1EE89 },
	{ 0x1EE8B, 0x1EE9B },
	{ 0x1EEA1, 0x1EEA3 },
	{ 0x1EEA5, 0x1EEA9 },
	{ 0x1EEAB, 0x1EEBB },
	{ 0x2F800, 0x2FA1D },
};

static const uint32_t alpha1[] = {
	0x00AA,
	0x00B5,
	0x00BA,
	0x0559,
	0x06FF,
	0x07B1,
	0x07FA,
	0x081A,
	0x0824,
	0x0828,
	0x093D,
	0x0950,
	0x09BD,
	0x09CE,
	0x09FC,
	0x0ABD,
	0x0AD0,
	0x0AF9,
	0x0B3D,
	0x0B71,
	0x0BD0,
	0x0C3D,
	0x0C5D,
	0x0C80,
	0x0CBD,
	0x0D3D,
	0x0D4E,
	0x0EBD,
	0x0F00,
	0x103F,
	0x1061,
	0x108E,
	0x10CD,
	0x17D7,
	0x17DC,
	0x1AA7,
	0x1CFA,
	0x2071,
	0x207F,
	0x2102,
	0x2107,
	0x214E,
	0x2D2D,
	0x2D6F,
	0x2E2F,
	0x3400,
	0x4DBF,
	0x4E00,
	0xA9CF,
	0xAA7A,
	0xAC00,
	0xD7A3,
	0x1083C,
	0x10A00,
	0x10F27,
	0x11075,
	0x11144,
	0x11147,
	0x11176,
	0x1133D,
	0x11350,
	0x11644,
	0x116B8,
	0x11909,
	0x11A00,
	0x11A3A,
	0x11A50,
	0x11A9D,
	0x11C40,
	0x11D46,
	0x11D98,
	0x11FB0,
	0x16F50,
	0x17000,
	0x187F7,
	0x18D00,
	0x18D08,
	0x1D4A2,
	0x1E14E,
	0x1E94B,
	0x1EE42,
	0x20000,
	0x2A6DF,
	0x2A700,
	0x2B738,
	0x2B740,
	0x2B81D,
	0x2B820,
	0x2CEA1,
	0x2CEB0,
	0x2EBE0,
	0x30000,
	0x3134A,
};

int
uni_isalpha(uint32_t r)
{
	const uint32_t *match;

	if ((match = bsearch(&r, alpha3, nelem(alpha3), sizeof *alpha3, &cmp2)))
		return !((r - match[0]) % 2);
	if (bsearch(&r, alpha2, nelem(alpha2), sizeof *alpha2, &cmp2))
		return 1;
	if (bsearch(&r, alpha1, nelem(alpha1), sizeof *alpha1, &cmp1))
		return 1;

	return 0;
}

static const uint32_t space2[][2] = {
	{ 0x0009, 0x000D },
	{ 0x001C, 0x0020 },
	{ 0x2000, 0x200A },
	{ 0x2028, 0x2029 },
};

static const uint32_t space1[] = {
	0x0085,
	0x00A0,
	0x1680,
	0x202F,
	0x205F,
	0x3000,
};

int
uni_isspace(uint32_t r)
{
	if (bsearch(&r, space2, nelem(space2), sizeof *space2, &cmp2))
		return 1;
	if (bsearch(&r, space1, nelem(space1), sizeof *space1, &cmp1))
		return 1;

	return 0;
}

static const uint32_t control2[][2] = {
	{ 0x0000, 0x001F },
	{ 0x007F, 0x009F },
};

int
uni_iscontrol(uint32_t r)
{
	if (bsearch(&r, control2, nelem(control2), sizeof *control2, &cmp2))
		return 1;

	return 0;
}

static const uint32_t upper3[][2] = {
	{ 0x0100, 0x012E },
	{ 0x0132, 0x0136 },
	{ 0x0139, 0x0147 },
	{ 0x014A, 0x0176 },
	{ 0x0179, 0x017D },
	{ 0x0182, 0x0184 },
	{ 0x01A0, 0x01A4 },
	{ 0x01B3, 0x01B5 },
	{ 0x01CD, 0x01DB },
	{ 0x01DE, 0x01EE },
	{ 0x01F8, 0x021E },
	{ 0x0222, 0x0232 },
	{ 0x0246, 0x024E },
	{ 0x0370, 0x0372 },
	{ 0x03D8, 0x03EE },
	{ 0x0460, 0x0480 },
	{ 0x048A, 0x04BE },
	{ 0x04C1, 0x04CD },
	{ 0x04D0, 0x052E },
	{ 0x1E00, 0x1E94 },
	{ 0x1EA0, 0x1EFE },
	{ 0x2C67, 0x2C6B },
	{ 0x2C80, 0x2CE2 },
	{ 0x2CEB, 0x2CED },
	{ 0xA640, 0xA66C },
	{ 0xA680, 0xA69A },
	{ 0xA722, 0xA72E },
	{ 0xA732, 0xA76E },
	{ 0xA779, 0xA77B },
	{ 0xA77E, 0xA786 },
	{ 0xA790, 0xA792 },
	{ 0xA796, 0xA7A8 },
	{ 0xA7B4, 0xA7C2 },
	{ 0xA7C7, 0xA7C9 },
	{ 0xA7D6, 0xA7D8 },
};

static const uint32_t upper2[][3] = {
	{ 0x0041, 0x005A, 0x0061 },
	{ 0x00C0, 0x00D6, 0x00E0 },
	{ 0x00D8, 0x00DE, 0x00F8 },
	{ 0x0189, 0x018A, 0x0256 },
	{ 0x01B1, 0x01B2, 0x028A },
	{ 0x0388, 0x038A, 0x03AD },
	{ 0x038E, 0x038F, 0x03CD },
	{ 0x0391, 0x03A1, 0x03B1 },
	{ 0x03A3, 0x03AB, 0x03C3 },
	{ 0x03D2, 0x03D4, 0x03D2 },
	{ 0x03FD, 0x03FF, 0x037B },
	{ 0x0400, 0x040F, 0x0450 },
	{ 0x0410, 0x042F, 0x0430 },
	{ 0x0531, 0x0556, 0x0561 },
	{ 0x10A0, 0x10C5, 0x2D00 },
	{ 0x13A0, 0x13EF, 0xAB70 },
	{ 0x13F0, 0x13F5, 0x13F8 },
	{ 0x1C90, 0x1CBA, 0x10D0 },
	{ 0x1CBD, 0x1CBF, 0x10FD },
	{ 0x1F08, 0x1F0F, 0x1F00 },
	{ 0x1F18, 0x1F1D, 0x1F10 },
	{ 0x1F28, 0x1F2F, 0x1F20 },
	{ 0x1F38, 0x1F3F, 0x1F30 },
	{ 0x1F48, 0x1F4D, 0x1F40 },
	{ 0x1F68, 0x1F6F, 0x1F60 },
	{ 0x1FB8, 0x1FB9, 0x1FB0 },
	{ 0x1FBA, 0x1FBB, 0x1F70 },
	{ 0x1FC8, 0x1FCB, 0x1F72 },
	{ 0x1FD8, 0x1FD9, 0x1FD0 },
	{ 0x1FDA, 0x1FDB, 0x1F76 },
	{ 0x1FE8, 0x1FE9, 0x1FE0 },
	{ 0x1FEA, 0x1FEB, 0x1F7A },
	{ 0x1FF8, 0x1FF9, 0x1F78 },
	{ 0x1FFA, 0x1FFB, 0x1F7C },
	{ 0x210B, 0x210D, 0x210B },
	{ 0x2110, 0x2112, 0x2110 },
	{ 0x2119, 0x211D, 0x2119 },
	{ 0x212C, 0x212D, 0x212C },
	{ 0x2130, 0x2131, 0x2130 },
	{ 0x213E, 0x213F, 0x213E },
	{ 0x2C00, 0x2C2F, 0x2C30 },
	{ 0x2C7E, 0x2C7F, 0x023F },
	{ 0xFF21, 0xFF3A, 0xFF41 },
	{ 0x10400, 0x10427, 0x10428 },
	{ 0x104B0, 0x104D3, 0x104D8 },
	{ 0x10570, 0x1057A, 0x10597 },
	{ 0x1057C, 0x1058A, 0x105A3 },
	{ 0x1058C, 0x10592, 0x105B3 },
	{ 0x10594, 0x10595, 0x105BB },
	{ 0x10C80, 0x10CB2, 0x10CC0 },
	{ 0x118A0, 0x118BF, 0x118C0 },
	{ 0x16E40, 0x16E5F, 0x16E60 },
	{ 0x1D400, 0x1D419, 0x1D400 },
	{ 0x1D434, 0x1D44D, 0x1D434 },
	{ 0x1D468, 0x1D481, 0x1D468 },
	{ 0x1D49E, 0x1D49F, 0x1D49E },
	{ 0x1D4A5, 0x1D4A6, 0x1D4A5 },
	{ 0x1D4A9, 0x1D4AC, 0x1D4A9 },
	{ 0x1D4AE, 0x1D4B5, 0x1D4AE },
	{ 0x1D4D0, 0x1D4E9, 0x1D4D0 },
	{ 0x1D504, 0x1D505, 0x1D504 },
	{ 0x1D507, 0x1D50A, 0x1D507 },
	{ 0x1D50D, 0x1D514, 0x1D50D },
	{ 0x1D516, 0x1D51C, 0x1D516 },
	{ 0x1D538, 0x1D539, 0x1D538 },
	{ 0x1D53B, 0x1D53E, 0x1D53B },
	{ 0x1D540, 0x1D544, 0x1D540 },
	{ 0x1D54A, 0x1D550, 0x1D54A },
	{ 0x1D56C, 0x1D585, 0x1D56C },
	{ 0x1D5A0, 0x1D5B9, 0x1D5A0 },
	{ 0x1D5D4, 0x1D5ED, 0x1D5D4 },
	{ 0x1D608, 0x1D621, 0x1D608 },
	{ 0x1D63C, 0x1D655, 0x1D63C },
	{ 0x1D670, 0x1D689, 0x1D670 },
	{ 0x1D6A8, 0x1D6C0, 0x1D6A8 },
	{ 0x1D6E2, 0x1D6FA, 0x1D6E2 },
	{ 0x1D71C, 0x1D734, 0x1D71C },
	{ 0x1D756, 0x1D76E, 0x1D756 },
	{ 0x1D790, 0x1D7A8, 0x1D790 },
	{ 0x1E900, 0x1E921, 0x1E922 },
};

static const uint32_t upper1[][2] = {
	{ 0x0130, 0x0069 },
	{ 0x0178, 0x00FF },
	{ 0x0181, 0x0253 },
	{ 0x0186, 0x0254 },
	{ 0x0187, 0x0188 },
	{ 0x018B, 0x018C },
	{ 0x018E, 0x01DD },
	{ 0x018F, 0x0259 },
	{ 0x0190, 0x025B },
	{ 0x0191, 0x0192 },
	{ 0x0193, 0x0260 },
	{ 0x0194, 0x0263 },
	{ 0x0196, 0x0269 },
	{ 0x0197, 0x0268 },
	{ 0x0198, 0x0199 },
	{ 0x019C, 0x026F },
	{ 0x019D, 0x0272 },
	{ 0x019F, 0x0275 },
	{ 0x01A6, 0x0280 },
	{ 0x01A7, 0x01A8 },
	{ 0x01A9, 0x0283 },
	{ 0x01AC, 0x01AD },
	{ 0x01AE, 0x0288 },
	{ 0x01AF, 0x01B0 },
	{ 0x01B7, 0x0292 },
	{ 0x01B8, 0x01B9 },
	{ 0x01BC, 0x01BD },
	{ 0x01C4, 0x01C6 },
	{ 0x01C7, 0x01C9 },
	{ 0x01CA, 0x01CC },
	{ 0x01F1, 0x01F3 },
	{ 0x01F4, 0x01F5 },
	{ 0x01F6, 0x0195 },
	{ 0x01F7, 0x01BF },
	{ 0x0220, 0x019E },
	{ 0x023A, 0x2C65 },
	{ 0x023B, 0x023C },
	{ 0x023D, 0x019A },
	{ 0x023E, 0x2C66 },
	{ 0x0241, 0x0242 },
	{ 0x0243, 0x0180 },
	{ 0x0244, 0x0289 },
	{ 0x0245, 0x028C },
	{ 0x0376, 0x0377 },
	{ 0x037F, 0x03F3 },
	{ 0x0386, 0x03AC },
	{ 0x038C, 0x03CC },
	{ 0x03CF, 0x03D7 },
	{ 0x03F4, 0x03B8 },
	{ 0x03F7, 0x03F8 },
	{ 0x03F9, 0x03F2 },
	{ 0x03FA, 0x03FB },
	{ 0x04C0, 0x04CF },
	{ 0x10C7, 0x2D27 },
	{ 0x10CD, 0x2D2D },
	{ 0x1E9E, 0x00DF },
	{ 0x1F59, 0x1F51 },
	{ 0x1F5B, 0x1F53 },
	{ 0x1F5D, 0x1F55 },
	{ 0x1F5F, 0x1F57 },
	{ 0x1FEC, 0x1FE5 },
	{ 0x2102, 0x2102 },
	{ 0x2107, 0x2107 },
	{ 0x2115, 0x2115 },
	{ 0x2124, 0x2124 },
	{ 0x2126, 0x03C9 },
	{ 0x2128, 0x2128 },
	{ 0x212A, 0x006B },
	{ 0x212B, 0x00E5 },
	{ 0x2132, 0x214E },
	{ 0x2133, 0x2133 },
	{ 0x2145, 0x2145 },
	{ 0x2183, 0x2184 },
	{ 0x2C60, 0x2C61 },
	{ 0x2C62, 0x026B },
	{ 0x2C63, 0x1D7D },
	{ 0x2C64, 0x027D },
	{ 0x2C6D, 0x0251 },
	{ 0x2C6E, 0x0271 },
	{ 0x2C6F, 0x0250 },
	{ 0x2C70, 0x0252 },
	{ 0x2C72, 0x2C73 },
	{ 0x2C75, 0x2C76 },
	{ 0x2CF2, 0x2CF3 },
	{ 0xA77D, 0x1D79 },
	{ 0xA78B, 0xA78C },
	{ 0xA78D, 0x0265 },
	{ 0xA7AA, 0x0266 },
	{ 0xA7AB, 0x025C },
	{ 0xA7AC, 0x0261 },
	{ 0xA7AD, 0x026C },
	{ 0xA7AE, 0x026A },
	{ 0xA7B0, 0x029E },
	{ 0xA7B1, 0x0287 },
	{ 0xA7B2, 0x029D },
	{ 0xA7B3, 0xAB53 },
	{ 0xA7C4, 0xA794 },
	{ 0xA7C5, 0x0282 },
	{ 0xA7C6, 0x1D8E },
	{ 0xA7D0, 0xA7D1 },
	{ 0xA7F5, 0xA7F6 },
	{ 0x1D49C, 0x1D49C },
	{ 0x1D4A2, 0x1D4A2 },
	{ 0x1D546, 0x1D546 },
	{ 0x1D7CA, 0x1D7CA },
};

int
uni_isupper(uint32_t r)
{
	const uint32_t *match;

	if ((match = bsearch(&r, upper3, nelem(upper3), sizeof *upper3, &cmp2)))
		return !((r - match[0]) % 2);
	if (bsearch(&r, upper2, nelem(upper2), sizeof *upper2, &cmp2))
		return 1;
	if (bsearch(&r, upper1, nelem(upper1), sizeof *upper1, &cmp1))
		return 1;

	return 0;
}

uint32_t
uni_tolower(uint32_t r)
{
	uint32_t *match;

	if ((match = bsearch(&r, upper3, nelem(upper3), sizeof *upper3, &cmp2)))

		return ((r - match[0]) % 2) ? r : r + 1;
	if ((match = bsearch(&r, upper2, nelem(upper2), sizeof *upper2, &cmp2)))

		return match[2] + (r - match[0]);
	if ((match = bsearch(&r, upper1, nelem(upper1), sizeof *upper1, &cmp1)))

		return match[1];

	return r;
}

static const uint32_t lower4[][2] = {
	{ 0x0101, 0x012F },
	{ 0x0133, 0x0137 },
	{ 0x013A, 0x0148 },
	{ 0x014B, 0x0177 },
	{ 0x017A, 0x017E },
	{ 0x0183, 0x0185 },
	{ 0x01A1, 0x01A5 },
	{ 0x01B4, 0x01B6 },
	{ 0x01CE, 0x01DC },
	{ 0x01DF, 0x01EF },
	{ 0x01F9, 0x021F },
	{ 0x0223, 0x0233 },
	{ 0x0247, 0x024F },
	{ 0x0371, 0x0373 },
	{ 0x03D9, 0x03EF },
	{ 0x0461, 0x0481 },
	{ 0x048B, 0x04BF },
	{ 0x04C2, 0x04CE },
	{ 0x04D1, 0x052F },
	{ 0x1E01, 0x1E95 },
	{ 0x1EA1, 0x1EFF },
	{ 0x2C68, 0x2C6C },
	{ 0x2C81, 0x2CE3 },
	{ 0x2CEC, 0x2CEE },
	{ 0xA641, 0xA66D },
	{ 0xA681, 0xA69B },
	{ 0xA723, 0xA72F },
	{ 0xA733, 0xA76F },
	{ 0xA77A, 0xA77C },
	{ 0xA77F, 0xA787 },
	{ 0xA791, 0xA793 },
	{ 0xA797, 0xA7A9 },
	{ 0xA7B5, 0xA7C3 },
	{ 0xA7C8, 0xA7CA },
	{ 0xA7D7, 0xA7D9 },
};

static const uint32_t lower2[][3] = {
	{ 0x0061, 0x007A, 0x0041 },
	{ 0x00E0, 0x00F6, 0x00C0 },
	{ 0x00F8, 0x00FE, 0x00D8 },
	{ 0x01AA, 0x01AB, 0x01AA },
	{ 0x0234, 0x0239, 0x0234 },
	{ 0x023F, 0x0240, 0x2C7E },
	{ 0x0256, 0x0257, 0x0189 },
	{ 0x025D, 0x025F, 0x025D },
	{ 0x026D, 0x026E, 0x026D },
	{ 0x0273, 0x0274, 0x0273 },
	{ 0x0276, 0x027C, 0x0276 },
	{ 0x027E, 0x027F, 0x027E },
	{ 0x0284, 0x0286, 0x0284 },
	{ 0x028A, 0x028B, 0x01B1 },
	{ 0x028D, 0x0291, 0x028D },
	{ 0x0295, 0x029C, 0x0295 },
	{ 0x029F, 0x02AF, 0x029F },
	{ 0x037B, 0x037D, 0x03FD },
	{ 0x03AD, 0x03AF, 0x0388 },
	{ 0x03B1, 0x03C1, 0x0391 },
	{ 0x03C3, 0x03CB, 0x03A3 },
	{ 0x03CD, 0x03CE, 0x038E },
	{ 0x0430, 0x044F, 0x0410 },
	{ 0x0450, 0x045F, 0x0400 },
	{ 0x0561, 0x0586, 0x0531 },
	{ 0x0587, 0x0588, 0x0587 },
	{ 0x10D0, 0x10FA, 0x1C90 },
	{ 0x10FD, 0x10FF, 0x1CBD },
	{ 0x13F8, 0x13FD, 0x13F0 },
	{ 0x1C83, 0x1C84, 0x0421 },
	{ 0x1D00, 0x1D2B, 0x1D00 },
	{ 0x1D6B, 0x1D77, 0x1D6B },
	{ 0x1D7A, 0x1D7C, 0x1D7A },
	{ 0x1D7E, 0x1D8D, 0x1D7E },
	{ 0x1D8F, 0x1D9A, 0x1D8F },
	{ 0x1E96, 0x1E9A, 0x1E96 },
	{ 0x1E9C, 0x1E9D, 0x1E9C },
	{ 0x1F00, 0x1F07, 0x1F08 },
	{ 0x1F10, 0x1F15, 0x1F18 },
	{ 0x1F20, 0x1F27, 0x1F28 },
	{ 0x1F30, 0x1F37, 0x1F38 },
	{ 0x1F40, 0x1F45, 0x1F48 },
	{ 0x1F60, 0x1F67, 0x1F68 },
	{ 0x1F70, 0x1F71, 0x1FBA },
	{ 0x1F72, 0x1F75, 0x1FC8 },
	{ 0x1F76, 0x1F77, 0x1FDA },
	{ 0x1F78, 0x1F79, 0x1FF8 },
	{ 0x1F7A, 0x1F7B, 0x1FEA },
	{ 0x1F7C, 0x1F7D, 0x1FFA },
	{ 0x1F80, 0x1F87, 0x1F88 },
	{ 0x1F90, 0x1F97, 0x1F98 },
	{ 0x1FA0, 0x1FA7, 0x1FA8 },
	{ 0x1FB0, 0x1FB1, 0x1FB8 },
	{ 0x1FB6, 0x1FB7, 0x1FB6 },
	{ 0x1FC6, 0x1FC7, 0x1FC6 },
	{ 0x1FD0, 0x1FD1, 0x1FD8 },
	{ 0x1FD2, 0x1FD3, 0x1FD2 },
	{ 0x1FD6, 0x1FD7, 0x1FD6 },
	{ 0x1FE0, 0x1FE1, 0x1FE8 },
	{ 0x1FE2, 0x1FE4, 0x1FE2 },
	{ 0x1FE6, 0x1FE7, 0x1FE6 },
	{ 0x1FF6, 0x1FF7, 0x1FF6 },
	{ 0x210E, 0x210F, 0x210E },
	{ 0x213C, 0x213D, 0x213C },
	{ 0x2146, 0x2149, 0x2146 },
	{ 0x2C30, 0x2C5F, 0x2C00 },
	{ 0x2C77, 0x2C7B, 0x2C77 },
	{ 0x2D00, 0x2D25, 0x10A0 },
	{ 0xA730, 0xA731, 0xA730 },
	{ 0xA771, 0xA778, 0xA771 },
	{ 0xAB30, 0xAB52, 0xAB30 },
	{ 0xAB54, 0xAB5A, 0xAB54 },
	{ 0xAB60, 0xAB68, 0xAB60 },
	{ 0xAB70, 0xABBF, 0x13A0 },
	{ 0xFB00, 0xFB06, 0xFB00 },
	{ 0xFB13, 0xFB17, 0xFB13 },
	{ 0xFF41, 0xFF5A, 0xFF21 },
	{ 0x10428, 0x1044F, 0x10400 },
	{ 0x104D8, 0x104FB, 0x104B0 },
	{ 0x10597, 0x105A1, 0x10570 },
	{ 0x105A3, 0x105B1, 0x1057C },
	{ 0x105B3, 0x105B9, 0x1058C },
	{ 0x105BB, 0x105BC, 0x10594 },
	{ 0x10CC0, 0x10CF2, 0x10C80 },
	{ 0x118C0, 0x118DF, 0x118A0 },
	{ 0x16E60, 0x16E7F, 0x16E40 },
	{ 0x1D41A, 0x1D433, 0x1D41A },
	{ 0x1D44E, 0x1D454, 0x1D44E },
	{ 0x1D456, 0x1D467, 0x1D456 },
	{ 0x1D482, 0x1D49B, 0x1D482 },
	{ 0x1D4B6, 0x1D4B9, 0x1D4B6 },
	{ 0x1D4BD, 0x1D4C3, 0x1D4BD },
	{ 0x1D4C5, 0x1D4CF, 0x1D4C5 },
	{ 0x1D4EA, 0x1D503, 0x1D4EA },
	{ 0x1D51E, 0x1D537, 0x1D51E },
	{ 0x1D552, 0x1D56B, 0x1D552 },
	{ 0x1D586, 0x1D59F, 0x1D586 },
	{ 0x1D5BA, 0x1D5D3, 0x1D5BA },
	{ 0x1D5EE, 0x1D607, 0x1D5EE },
	{ 0x1D622, 0x1D63B, 0x1D622 },
	{ 0x1D656, 0x1D66F, 0x1D656 },
	{ 0x1D68A, 0x1D6A5, 0x1D68A },
	{ 0x1D6C2, 0x1D6DA, 0x1D6C2 },
	{ 0x1D6DC, 0x1D6E1, 0x1D6DC },
	{ 0x1D6FC, 0x1D714, 0x1D6FC },
	{ 0x1D716, 0x1D71B, 0x1D716 },
	{ 0x1D736, 0x1D74E, 0x1D736 },
	{ 0x1D750, 0x1D755, 0x1D750 },
	{ 0x1D770, 0x1D788, 0x1D770 },
	{ 0x1D78A, 0x1D78F, 0x1D78A },
	{ 0x1D7AA, 0x1D7C2, 0x1D7AA },
	{ 0x1D7C4, 0x1D7C9, 0x1D7C4 },
	{ 0x1DF00, 0x1DF09, 0x1DF00 },
	{ 0x1DF0B, 0x1DF1E, 0x1DF0B },
	{ 0x1E922, 0x1E943, 0x1E900 },
};

static const uint32_t lower1[][2] = {
	{ 0x00B5, 0x039C },
	{ 0x00DF, 0x00DF },
	{ 0x00FF, 0x0178 },
	{ 0x0131, 0x0049 },
	{ 0x0138, 0x0138 },
	{ 0x0149, 0x0149 },
	{ 0x017F, 0x0053 },
	{ 0x0180, 0x0243 },
	{ 0x0188, 0x0187 },
	{ 0x018C, 0x018B },
	{ 0x018D, 0x018D },
	{ 0x0192, 0x0191 },
	{ 0x0195, 0x01F6 },
	{ 0x0199, 0x0198 },
	{ 0x019A, 0x023D },
	{ 0x019B, 0x019B },
	{ 0x019E, 0x0220 },
	{ 0x01A8, 0x01A7 },
	{ 0x01AD, 0x01AC },
	{ 0x01B0, 0x01AF },
	{ 0x01B9, 0x01B8 },
	{ 0x01BA, 0x01BA },
	{ 0x01BD, 0x01BC },
	{ 0x01BE, 0x01BE },
	{ 0x01BF, 0x01F7 },
	{ 0x01C6, 0x01C4 },
	{ 0x01C9, 0x01C7 },
	{ 0x01CC, 0x01CA },
	{ 0x01DD, 0x018E },
	{ 0x01F0, 0x01F0 },
	{ 0x01F3, 0x01F1 },
	{ 0x01F5, 0x01F4 },
	{ 0x0221, 0x0221 },
	{ 0x023C, 0x023B },
	{ 0x0242, 0x0241 },
	{ 0x0250, 0x2C6F },
	{ 0x0251, 0x2C6D },
	{ 0x0252, 0x2C70 },
	{ 0x0253, 0x0181 },
	{ 0x0254, 0x0186 },
	{ 0x0255, 0x0255 },
	{ 0x0258, 0x0258 },
	{ 0x0259, 0x018F },
	{ 0x025A, 0x025A },
	{ 0x025B, 0x0190 },
	{ 0x025C, 0xA7AB },
	{ 0x0260, 0x0193 },
	{ 0x0261, 0xA7AC },
	{ 0x0262, 0x0262 },
	{ 0x0263, 0x0194 },
	{ 0x0264, 0x0264 },
	{ 0x0265, 0xA78D },
	{ 0x0266, 0xA7AA },
	{ 0x0267, 0x0267 },
	{ 0x0268, 0x0197 },
	{ 0x0269, 0x0196 },
	{ 0x026A, 0xA7AE },
	{ 0x026B, 0x2C62 },
	{ 0x026C, 0xA7AD },
	{ 0x026F, 0x019C },
	{ 0x0270, 0x0270 },
	{ 0x0271, 0x2C6E },
	{ 0x0272, 0x019D },
	{ 0x0275, 0x019F },
	{ 0x027D, 0x2C64 },
	{ 0x0280, 0x01A6 },
	{ 0x0281, 0x0281 },
	{ 0x0282, 0xA7C5 },
	{ 0x0283, 0x01A9 },
	{ 0x0287, 0xA7B1 },
	{ 0x0288, 0x01AE },
	{ 0x0289, 0x0244 },
	{ 0x028C, 0x0245 },
	{ 0x0292, 0x01B7 },
	{ 0x0293, 0x0293 },
	{ 0x029D, 0xA7B2 },
	{ 0x029E, 0xA7B0 },
	{ 0x0377, 0x0376 },
	{ 0x0390, 0x0390 },
	{ 0x03AC, 0x0386 },
	{ 0x03B0, 0x03B0 },
	{ 0x03C2, 0x03A3 },
	{ 0x03CC, 0x038C },
	{ 0x03D0, 0x0392 },
	{ 0x03D1, 0x0398 },
	{ 0x03D5, 0x03A6 },
	{ 0x03D6, 0x03A0 },
	{ 0x03D7, 0x03CF },
	{ 0x03F0, 0x039A },
	{ 0x03F1, 0x03A1 },
	{ 0x03F2, 0x03F9 },
	{ 0x03F3, 0x037F },
	{ 0x03F5, 0x0395 },
	{ 0x03F8, 0x03F7 },
	{ 0x03FB, 0x03FA },
	{ 0x03FC, 0x03FC },
	{ 0x04CF, 0x04C0 },
	{ 0x0560, 0x0560 },
	{ 0x1C80, 0x0412 },
	{ 0x1C81, 0x0414 },
	{ 0x1C82, 0x041E },
	{ 0x1C85, 0x0422 },
	{ 0x1C86, 0x042A },
	{ 0x1C87, 0x0462 },
	{ 0x1C88, 0xA64A },
	{ 0x1D79, 0xA77D },
	{ 0x1D7D, 0x2C63 },
	{ 0x1D8E, 0xA7C6 },
	{ 0x1E9B, 0x1E60 },
	{ 0x1E9F, 0x1E9F },
	{ 0x1F50, 0x1F50 },
	{ 0x1F51, 0x1F59 },
	{ 0x1F52, 0x1F52 },
	{ 0x1F53, 0x1F5B },
	{ 0x1F54, 0x1F54 },
	{ 0x1F55, 0x1F5D },
	{ 0x1F56, 0x1F56 },
	{ 0x1F57, 0x1F5F },
	{ 0x1FB2, 0x1FB2 },
	{ 0x1FB3, 0x1FBC },
	{ 0x1FB4, 0x1FB4 },
	{ 0x1FBE, 0x0399 },
	{ 0x1FC2, 0x1FC2 },
	{ 0x1FC3, 0x1FCC },
	{ 0x1FC4, 0x1FC4 },
	{ 0x1FE5, 0x1FEC },
	{ 0x1FF2, 0x1FF2 },
	{ 0x1FF3, 0x1FFC },
	{ 0x1FF4, 0x1FF4 },
	{ 0x210A, 0x210A },
	{ 0x2113, 0x2113 },
	{ 0x212F, 0x212F },
	{ 0x2134, 0x2134 },
	{ 0x2139, 0x2139 },
	{ 0x214E, 0x2132 },
	{ 0x2184, 0x2183 },
	{ 0x2C61, 0x2C60 },
	{ 0x2C65, 0x023A },
	{ 0x2C66, 0x023E },
	{ 0x2C71, 0x2C71 },
	{ 0x2C73, 0x2C72 },
	{ 0x2C74, 0x2C74 },
	{ 0x2C76, 0x2C75 },
	{ 0x2CE4, 0x2CE4 },
	{ 0x2CF3, 0x2CF2 },
	{ 0x2D27, 0x10C7 },
	{ 0x2D2D, 0x10CD },
	{ 0xA78C, 0xA78B },
	{ 0xA78E, 0xA78E },
	{ 0xA794, 0xA7C4 },
	{ 0xA795, 0xA795 },
	{ 0xA7AF, 0xA7AF },
	{ 0xA7D1, 0xA7D0 },
	{ 0xA7D3, 0xA7D3 },
	{ 0xA7D5, 0xA7D5 },
	{ 0xA7F6, 0xA7F5 },
	{ 0xA7FA, 0xA7FA },
	{ 0xAB53, 0xA7B3 },
	{ 0x1D4BB, 0x1D4BB },
	{ 0x1D7CB, 0x1D7CB },
};

int
uni_islower(uint32_t r)
{
	const uint32_t *match;

	if ((match = bsearch(&r, lower4, nelem(lower4), sizeof *lower4, &cmp2)))
		return !((r - match[0]) % 2);
	if (bsearch(&r, lower2, nelem(lower2), sizeof *lower2, &cmp2))
		return 1;
	if (bsearch(&r, lower1, nelem(lower1), sizeof *lower1, &cmp1))
		return 1;

	return 0;
}

uint32_t
uni_toupper(uint32_t r)
{
	uint32_t *match;

	if ((match = bsearch(&r, lower4, nelem(lower4), sizeof *lower4, &cmp2)))

		return ((r - match[0]) % 2) ? r : r - 1;
	if ((match = bsearch(&r, lower2, nelem(lower2), sizeof *lower2, &cmp2)))

		return match[2] + (r - match[0]);
	if ((match = bsearch(&r, lower1, nelem(lower1), sizeof *lower1, &cmp1)))

		return match[1];

	return r;
}

static const uint32_t title2[][2] = {
	{ 0x1F88, 0x1F8F },
	{ 0x1F98, 0x1F9F },
	{ 0x1FA8, 0x1FAF },
};

static const uint32_t title1[] = {
	0x01C5,
	0x01C8,
	0x01CB,
	0x01F2,
	0x1FBC,
	0x1FCC,
	0x1FFC,
};

int
uni_istitle(uint32_t r)
{
	if (bsearch(&r, title2, nelem(title2), sizeof *title2, &cmp2))
		return 1;
	if (bsearch(&r, title1, nelem(title1), sizeof *title1, &cmp1))
		return 1;

	return 0;
}

static const uint32_t digit2[][2] = {
	{ 0x0030, 0x0039 },
	{ 0x0660, 0x0669 },
	{ 0x06F0, 0x06F9 },
	{ 0x07C0, 0x07C9 },
	{ 0x0966, 0x096F },
	{ 0x09E6, 0x09EF },
	{ 0x0A66, 0x0A6F },
	{ 0x0AE6, 0x0AEF },
	{ 0x0B66, 0x0B6F },
	{ 0x0BE6, 0x0BEF },
	{ 0x0C66, 0x0C6F },
	{ 0x0CE6, 0x0CEF },
	{ 0x0D66, 0x0D6F },
	{ 0x0DE6, 0x0DEF },
	{ 0x0E50, 0x0E59 },
	{ 0x0ED0, 0x0ED9 },
	{ 0x0F20, 0x0F29 },
	{ 0x1040, 0x1049 },
	{ 0x1090, 0x1099 },
	{ 0x17E0, 0x17E9 },
	{ 0x1810, 0x1819 },
	{ 0x1946, 0x194F },
	{ 0x19D0, 0x19D9 },
	{ 0x1A80, 0x1A89 },
	{ 0x1A90, 0x1A99 },
	{ 0x1B50, 0x1B59 },
	{ 0x1BB0, 0x1BB9 },
	{ 0x1C40, 0x1C49 },
	{ 0x1C50, 0x1C59 },
	{ 0xA620, 0xA629 },
	{ 0xA8D0, 0xA8D9 },
	{ 0xA900, 0xA909 },
	{ 0xA9D0, 0xA9D9 },
	{ 0xA9F0, 0xA9F9 },
	{ 0xAA50, 0xAA59 },
	{ 0xABF0, 0xABF9 },
	{ 0xFF10, 0xFF19 },
	{ 0x104A0, 0x104A9 },
	{ 0x10D30, 0x10D39 },
	{ 0x11066, 0x1106F },
	{ 0x110F0, 0x110F9 },
	{ 0x11136, 0x1113F },
	{ 0x111D0, 0x111D9 },
	{ 0x112F0, 0x112F9 },
	{ 0x11450, 0x11459 },
	{ 0x114D0, 0x114D9 },
	{ 0x11650, 0x11659 },
	{ 0x116C0, 0x116C9 },
	{ 0x11730, 0x11739 },
	{ 0x118E0, 0x118E9 },
	{ 0x11950, 0x11959 },
	{ 0x11C50, 0x11C59 },
	{ 0x11D50, 0x11D59 },
	{ 0x11DA0, 0x11DA9 },
	{ 0x16A60, 0x16A69 },
	{ 0x16AC0, 0x16AC9 },
	{ 0x16B50, 0x16B59 },
	{ 0x1D7CE, 0x1D7FF },
	{ 0x1E140, 0x1E149 },
	{ 0x1E2F0, 0x1E2F9 },
	{ 0x1E950, 0x1E959 },
	{ 0x1FBF0, 0x1FBF9 },
};

int
uni_isdigit(uint32_t r)
{
	if (bsearch(&r, digit2, nelem(digit2), sizeof *digit2, &cmp2))
		return 1;

	return 0;
}

size_t
uni8_encode(uint8_t dst[], size_t dstsz, uint32_t point)
{
	assert(dst);

	size_t written;

	switch ((written = uni32_sizeof(point))) {
	case 1:
		if (dstsz < 1)
			goto erange;

		dst[0] = (uint8_t)point;
		break;
	case 2:
		if (dstsz < 2)
			goto erange;

		dst[0] = 0xC0 | ((point >> 6)  & 0x1F);
		dst[1] = 0x80 | (point & 0x3F);
		break;
	case 3:
		if (dstsz < 3)
			goto erange;

		dst[0] = 0xE0 | ((point >> 12) & 0xF );
		dst[1] = 0x80 | ((point >> 6)  & 0x3F);
		dst[2] = 0x80 | (point & 0x3F);
		break;
	case 4:
		if (dstsz < 4)
			goto erange;

		dst[0] = 0xF0 | ((point >> 18) & 0x7 );
		dst[1] = 0x80 | ((point >> 12) & 0x3F);
		dst[2] = 0x80 | ((point >> 6)  & 0x3F);
		dst[3] = 0x80 | (point & 0x3F);
		break;
	default:
		break;
	}

	return written;

erange:
	errno = ERANGE;

	return -1;
}

size_t
uni8_decode(const uint8_t src[], uint32_t *point)
{
	assert(src);
	assert(point);

	size_t parsed;

	switch ((parsed = uni8_sizeof(*src))) {
	case 1:
		*point = src[0];
		break;
	case 2:
		if (!src[1])
			goto eilseq;

		*point =  (src[0] & 0x1f) << 6;
		*point |= (src[1] & 0x3f);
		break;
	case 3:
		if (!src[1] || !src[2])
			goto eilseq;

		*point =  (src[0] & 0x0f) << 12;
		*point |= (src[1] & 0x3f) << 6;
		*point |= (src[2] & 0x3f);
		break;
	case 4:
		if (!src[1] || !src[2] || !src[3])
			goto eilseq;

		*point =  (src[0] & 0x07) << 16;
		*point |= (src[1] & 0x3f) << 12;
		*point |= (src[2] & 0x3f) << 6;
		*point |= (src[3] & 0x3f);
		break;
	default:
		break;
	}

	return parsed;

eilseq:
	errno = EILSEQ;

	return -1;
}

size_t
uni8_sizeof(uint8_t c)
{
	if (c <= 127)
		return 1;
	if ((c & 0xE0) == 0xC0)
		return 2;
	if ((c & 0xF0) == 0xE0)
		return 3;
	if ((c & 0xF8) == 0xF0)
		return 4;

	errno = EILSEQ;
	return -1;
}

size_t
uni8_length(const uint8_t src[])
{
	assert(src);

	size_t total = 0, gap;

	while (*src) {
		if ((gap = uni8_sizeof(*src)) == (size_t)-1)
			return -1;

		total += 1;
		src += gap;
	}

	return total;
}

size_t
uni8_to32(const uint8_t src[], uint32_t dst[], size_t dstsz)
{
	assert(src);
	assert(dst);

	size_t nwritten = 0, gap;

	for (; *src && dstsz; --dstsz) {
		if ((gap = uni8_decode(src, dst++)) == (size_t)-1)
			return -1;

		src += gap;
		++nwritten;
	}

	/* No more space to store NUL. */
	if (dstsz == 0) {
		errno = ERANGE;
		return -1;
	}

	*dst = 0;

	return nwritten;
}

size_t
uni32_sizeof(uint32_t c)
{
	if (c <= 0x7F)
		return 1;
	if (c <= 0x7FF)
		return 2;
	if (c <= 0xFFFF)
		return 3;
	if (c <= 0x1FFFFF)
		return 4;

	errno = EILSEQ;
	return -1;
}

size_t
uni32_length(const uint32_t src[])
{
	assert(src);

	size_t total = 0;

	while (*src++)
		total++;

	return total;
}

size_t
uni32_requires(const uint32_t src[])
{
	assert(src);

	size_t total = 0, gap;

	while (*src) {
		if ((gap = uni32_sizeof(*src++)) == (size_t)-1)
			return -1;
		if (gap >= SIZE_MAX - total) {
			errno = ERANGE;
			return -1;
		}

		total += gap;
	}

	return total;
}

size_t
uni32_to8(const uint32_t src[], uint8_t dst[], size_t dstsz)
{
	assert(src);
	assert(dst);

	size_t nwritten = 0, gap;

	while (*src && dstsz) {
		if ((gap = uni8_encode(dst, dstsz, *src++)) == (size_t)-1)
			return -1;

		dst += gap;
		dstsz -= gap;
		nwritten += gap;
	}

	if (dstsz == 0) {
		errno = ERANGE;
		return -1;
	}

	*dst = 0;

	return nwritten;
}
